/* ===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================*/

#include <ncbi.h>
#include <sys/lock.h>
#include <gishlib.h>
#include "nrdb.h"

static void	compress PROTO((SeqStrPtr));
static void hashsave PROTO((SeqStrPtr sp, unsigned long hashval));

NRFilePtr
NRFile_Open(filename, dbname)
	char	*filename;
	char	*dbname;
{
	register NRFilePtr	nrfp;
	register FILE	*fp;
	static int	filenum = 0;

	fp = openfile(filename, "r");
	if (fp != NULL)
		setvbuf(fp, NULL, _IOFBF, 256*1024);
	else
		return NULL;

	nrfp = (NRFilePtr)mem_calloc(sizeof(*nrfp), 1);
	nrfp->fp = fp;
	nrfp->filename = str_dup(filename);
	nrfp->dbname = str_dup(dbname);
	nrfp->dbnamelen = strlen(dbname);
	nrfp->numseqs = 0;
	nrfp->nummatches = 0;
	nrfp->totres = 0;
	nrfp->dupres = 0;
	nrfp->numnull = 0;
	nrfp->filenum = ++filenum;
	return nrfp;
}


int
NRFile_Close(nrfp)
	NRFilePtr	nrfp;
{
	int		rc;

	rc = fclose(nrfp->fp);
	nrfp->fp = NULL;
	return rc;
}


void
NRFile_Destruct(nrfp)
	NRFilePtr	nrfp;
{

	fclose(nrfp->fp);
	nrfp->fp = NULL;
	nrfp->dbname[0] = nrfp->filename[0] = NULLB;
	nrfp->dbnamelen = 0;
	(void) free(nrfp);
	return;
}


Boolean
NRFile_Isopen(nrfp)
	NRFilePtr	nrfp;
{
	return (nrfp->fp != NULL);
}


SeqStrPtr
NRFile_Read(nrfp)
	NRFilePtr	nrfp;
{
	static SeqStr	ss;
	static SeqName	ssn;
	static size_t	namemax;
	FILE	*fp;
	register unsigned char	*cp;
	register unsigned char	ch;
	register unsigned long	hashval = 0;
	int	i;

	ss.chain = NULL;
	ss.clen = 0;
	ss.name1 = &ssn;
	ssn.chain = NULL;
	ssn.nrfp = nrfp;

	fp = nrfp->fp;
	if (getfasta(xltab, &ssn.name, &namemax, &ssn.namelen, &seqbuf, &seqbufmax, &ss.seqlen, fp) != 0)
		return NULL;

	maxnamelen = MAX(maxnamelen, ssn.namelen);
	
	ss.seq = (CharPtr)seqbuf;
	if (ss.seqlen > 0) {
		cp = (unsigned char *)ss.seq;
		while (ch = *cp++) {
			hashval *= 1103515245;
			hashval += (unsigned long)ch + 12345;
			if (!nttst[ch])
				break;
		}
		if (ch) {
			while (ch = *cp++) {
				hashval *= 1103515245;
				hashval += (unsigned long)ch + 12345;
			}
		}
		else
			compress(&ss);
	}

	ss.id = seqcnt;
	hashsave(&ss, hashval);

	nrfp->numseqs++;
	nrfp->totres += ss.seqlen;

	if (ss.seqlen > maxseqlen)
		maxseqlen = ss.seqlen;

	return &ss;
}

static void
compress(ssp)
	SeqStrPtr	ssp;
{
	register unsigned char	*cp, *ccp;
	register unsigned char	ch;
	unsigned char	*cpend, bch;

	cp = ccp = (unsigned char *)ssp->seq;
	cpend = ccp + ssp->seqlen / 4;

	if (ccp < cpend) {
		bch = (((( (nt_atob[cp[0]] << 2) |
				nt_atob[cp[1]]) << 2) | nt_atob[cp[2]]) << 2) |
				nt_atob[cp[3]];
		ssp->seq[0] = bch;
		cp += 4;
		while (++ccp < cpend) {
			*ccp = (((( (nt_atob[cp[0]] << 2) |
					nt_atob[cp[1]]) << 2) | nt_atob[cp[2]]) << 2) |
					nt_atob[cp[3]];
			cp += 4;
		}
	}

	if (ch = *cp++) {
		bch = nt_atob[ch] << 2;
		if (ch = *cp++) {
			bch |= nt_atob[ch];
			bch <<= 2;
			if (ch = *cp++)
				bch |= nt_atob[ch];
			bch <<= 2;
		}
		else
			bch <<= 4;
		*ccp++ = bch;
	}

	ssp->clen = ccp - (unsigned char *)ssp->seq;
}

static void
hashsave(sp, hashval)
	SeqStrPtr	sp;
	unsigned long hashval;
{
	register HashRayPtr	newhrp;

	if (sp->id >= hashraymax) {
		hashraymax += 10000;
		hashray = (HashRayPtr)mem_realloc(hashray, sizeof(*hashray)*hashraymax);
		if (hashray == NULL)
			fatal(1, "Out of memory for hashray");
	}
	hashray[sp->id].sp = sp;
	hashray[sp->id].hashval = hashval;
	hashray[sp->id].nextid = ULONG_MAX;
	return;
}
